﻿/*************************************************************************************
* HanQin(Caesar) Cai                                                                 *
* 2012 Saint Norbert College Computer Science Senior Capstone Project                *
* TEXT RECOGNITION                                                                   *
* This is a C# program using tessnet2, a open source OCR engine by Google,           *
* to achieve the text recognition                                                    *
* The development of this project is based on VS2010 and .NET 3.5                    *
* However, the program can run on .NET 4 with a configuration file                   *
* ************************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq; 
using System.Text;
using System.Windows.Forms;
using System.IO; 
using System.Runtime.InteropServices;
using System.Threading;

namespace Project
{
    public partial class Recognition : Form
    {
        /********************************************************************
         * Local Data                                                       *
         * ******************************************************************/
        /// <summary>
        /// How many noising points allowed in an empty line, usually 0.01-0.05
        /// </summary>
        private const double emptyLineCoefficient = 0.015;
        /// <summary>
        /// File filter for the pictures
        /// </summary>
        private string filter = "Picture File |*.jpg;*.tif;*.bmp";
        /// <summary>
        /// Default file extension
        /// </summary>
        private string fileExtension = "jpg";
        /// <summary>
        /// grapics
        /// </summary>
        private Graphics g = null;
        /// <summary>
        /// Pen for the box
        /// </summary>
        private Pen pen = new Pen(Color.Red, 2);
        /// <summary>
        /// Drawing the box while cutting the picture
        /// </summary>
        private bool isDraw = false;
        /// <summary>
        /// Cutting thepicture or not
        /// </summary>
        private bool isCut = false;
        /// <summary>
        /// Range the picture cut
        /// </summary>
        private Rectangle cutRect = new Rectangle();
        /// <summary>
        /// Picture name when load from file
        /// </summary>
        private string PicName;
        /// <summary>
        /// Process for the full recognition
        /// </summary>
        private int process = 0;
        /// <summary>
        /// Zoom in or Zoom out rate for the image in the picturebox
        /// </summary>
        private double zoomrate;
        /// <summary>
        /// Fit the Height or fit the Width
        /// </summary>
        private bool fitHeight;
        /// <summary>
        /// The image that will be recognized
        /// </summary>
        private Bitmap image;
        /// <summary>
        /// Where the image is located in the picturebox
        /// </summary>
        private Rectangle imgRect = new Rectangle();
        /// <summary>
        /// The text color user picked
        /// </summary>
        private Color textColor = Color.Black;
        /// <summary>
        /// Whether picking the text color
        /// </summary>
        private bool isPick = false;
        /// <summary>
        /// Record the Y coordinates between two lines, total 1000 lines allowed.
        /// </summary>
        private int[] lineCut = new int[1000];

        /********************************************************************
         * Form Constructor                                                 *
         * ******************************************************************/
        public Recognition()
        {
            InitializeComponent();
            tbOutput.Text = "Welcome!! \r\nClick \"Start Text Recognition\" button to load a JPG picture.";
            g = pbInput.CreateGraphics();
            // set styles to reduce flicker and painting over twice
            // source: http://www.gamedev.net/topic/447943-cgdi---any-way-to-make-this-faster/
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.UserPaint, true);
            this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);

            tsProgressBar.Minimum = 0;
            tsProgressBar.Maximum = 10000;
        }

        /********************************************************************
         * Button Events, the main process                                  *
         * ******************************************************************/
        private void btStart_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new System.Windows.Forms.OpenFileDialog();
            // Set up dialog
            dlg.DefaultExt = fileExtension;
            dlg.Filter = filter;

            // Get the file to open
            DialogResult res = dlg.ShowDialog();
            if (res == DialogResult.OK)
            {
                // Initialization for restart
                newToolStripButton.PerformClick();
                btStart.Text = "Restart";
                tbOutput.Text = "Cool!!\r\nNow, if you have a nice and clear text picture,  you can click \"Quick Recognition\" button, the program will then finish the recognition automaticly.\r\n(Notice that the Quick Recognition can only handle black text.)\r\nOtherwise, you may click \"Full Process\" button, the full process of recognition should be more accurate, but take more steps and time.";
                PicName = dlg.FileName;
                image = new Bitmap(PicName);
                pbInput.Image = image;
                convertSize();
            }
            else return;
        }
		
		//The Full Process includes selecting the text area, picking text color, binarization, denoising, and recognition
        private void btFull_Click(object sender, EventArgs e)
        {
            if (pbInput.Image == null)
            {
                MessageBox.Show("Please load a picture first.");
            }
            else
            {
                switch (process)
                {
                    case 0:           //Setup for cutting picture
                        btFull.Text = "Next >>";
                        btQuick.Enabled = false;
                        isCut = true;
                        tbOutput.Text = "You can drag the mouse on the picture to select the area that includes the text.\r\nWhen you finish, click \"Next\" button.";
                        process++;
                        break;
                    case 1:                  //Cutting Picture and setup for picking the text color 
                        if (!isCut)
                        {
                            if (fitHeight)
                                image = CropBitmap(image, (int)((cutRect.X - imgRect.X) / zoomrate), (int)(cutRect.Y / zoomrate), (int)(cutRect.Width / zoomrate), (int)(cutRect.Height / zoomrate));
                            else image = CropBitmap(image, (int)(cutRect.X / zoomrate), (int)((cutRect.Y - imgRect.Y) / zoomrate), (int)(cutRect.Width / zoomrate), (int)(cutRect.Height / zoomrate));
                            pbInput.Image = image;
                            pbInput.Refresh();
                            convertSize();
                            tbOutput.Text = "Ok, now please pick the color of the text by clicking the left mouse button on the picture. \r\nDon't pick a color too dark or too light, an average one will be nice.\r\nWhen you finish, click \"Next\" button.";
                            label1.Text = "Picked Color:";
                            labelColor.BackColor = textColor;
                            isPick = true;
                            process++;
                        }
                        else MessageBox.Show("You have to use mouse to select the text part of the picture.");
                        break;
                    case 2:            //Binarize the Picture by using the picked color and setup for denoising
                        Binarize();
                        isPick = false;
                        label1.Text = " ";
                        labelColor.BackColor = this.BackColor;
                        tbOutput.Text = "You can enter the maximum block size of denoising in the text box.\r\nThe block size should be an integer between 1 and 50.\r\nIf you don't know how many it should be, I suggest 10-20 for bold font, 2-6 for slim font.\r\nYou may enter 1, if you think this picture is clear enough.\r\nWhen you finish, click \"Next\" button.";
                        pbInput.Refresh();
                        label2.Text = "Block Size:";
                        tbBlockSize.Visible = true;
                        tbBlockSize.Text = "10";
                        process++;
                        break;
                    case 3:                  //Denoising and Recognition
                        if(tbBlockSize.Text == String.Empty)
                        {
                            MessageBox.Show("Block Size cannot be empty\r\nIf you don't know how many it should be, I suggest 10-20 for Bold font, 2-6 for Skinny font");
                            return;
                        }
                        else if (!IsNumeric(tbBlockSize.Text))
                        {
                            MessageBox.Show("Block Size must be number\r\nIf you don't know how many it should be, I suggest 10-20 for Bold font, 2-6 for Skinny font");
                            return;
                        }
                        else if (Convert.ToInt32(tbBlockSize.Text) < 1 || Convert.ToInt32(tbBlockSize.Text)>50)
                        {
                            MessageBox.Show("The Block Size should be between 1 and 50\r\nIf you don't know how many it should be, I suggest 10-20 for Bold font, 2-6 for Skinny font");
                            return;
                        }
                        Denoising(image, Convert.ToInt32(tbBlockSize.Text));
                        pbInput.Image = image;
                        RecognitionText();
                        label2.Text = " ";
                        tbBlockSize.Visible = false;
                        btFull.Enabled = false;
                        break;
                    default:
                        MessageBox.Show("Error, please reload a picture.");
                        btStart.PerformClick();
                        break;
                }
            }
        }

		//The Quick Process only includes binarization and recognition, where binaryization use the default Black color
        private void btQuick_Click(object sender, EventArgs e)
        {
            if (pbInput.Image == null)
            {
                MessageBox.Show("Please load a picture first.");
            }
            else
            {
                Binarize();
                //RemoveBlock(image, 4);
                RecognitionText();
            }
        }

        /********************************************************************
         * Tool Strip Events                                                *
         * ******************************************************************/
		 //Initialize everything
        private void newToolStripButton_Click(object sender, EventArgs e)
        {
            for (int i = 0; i < 1000; i++)
                lineCut[i] = -1;
            process = 0;
            tsLabel1.Text = "Ready! ";
            tsProgressBar.Value = 0;
            tsProgressBar.Visible = false;
            btStart.Text = "Start Text Recognition";
            btFull.Text = "Full Process >> (Recommended)";
            btFull.Enabled = true;
            btQuick.Enabled = true;
            isCut = false;
            isDraw = false;
            isPick = false;
            textColor = Color.Black;
            label1.Text = " ";
            labelColor.BackColor = this.BackColor;
            label2.Text = " ";
            tbBlockSize.Visible = false;
            tbOutput.Text = "Welcome!! \r\nClick \"Start Text Recognition\" button to load a picture.";
            pbInput.Image = null;
            Refresh();
        }

        private void openToolStripButton_Click(object sender, EventArgs e)
        {
            btStart.PerformClick();
        }

		//Help page
        private void helpToolStripButton_Click(object sender, EventArgs e)
        {
            Help form1 = new Help();
            form1.ShowDialog();
        }

        /*********************************************************************************
         * Compute relationship between the coordinate in picture box and original image *
         * *******************************************************************************/
        private void convertSize()
        {
            if (image.Height / image.Width >= pbInput.Height / pbInput.Width)
            {            //The picture's height fits the picture box
                zoomrate = (double)pbInput.Height / (double)image.Height;
                fitHeight = true;
                imgRect.Y = 0;
                imgRect.Height = pbInput.Height;
                imgRect.Width = (int)(zoomrate * image.Width);
                imgRect.X = (pbInput.Width - imgRect.Width) / 2;
            }
            else
            {            //The picture's height fits the picture box
                zoomrate = (double)pbInput.Width / (double)image.Width;
                fitHeight = false;
                imgRect.X = 0;
                imgRect.Width = pbInput.Width;
                imgRect.Height = (int)(zoomrate * image.Height);
                imgRect.Y = (pbInput.Height - imgRect.Height) / 2;
            }
        }

        /******************************************************************************************
         * Use Mouse Events to cut the picture and pick the text color                            *
         * Source for CropBitmap:http://www.nerdydork.com/crop-an-image-bitmap-in-c-or-vbnet.html *
         * ****************************************************************************************/
        private void pbInput_MouseDown(object sender, MouseEventArgs e)
        {
            if (isCut)
            {
                if (e.Button != MouseButtons.Left)
                    return;

                if (e.X < imgRect.Left || e.X > imgRect.Right || e.Y < imgRect.Top || e.Y > imgRect.Bottom)
                {
                    MessageBox.Show("You have to cut on the picture.");
                    return;
                }
                cutRect.X = e.X;
                cutRect.Y = e.Y;
                this.isDraw = true;
            }
        }

        private void pbInput_MouseMove(object sender, MouseEventArgs e)
        {
            //tsLabel1.Text = "x:" + e.X + "  y:" + e.Y;               //used for test
            if (isCut)
            {
                if (isDraw)
                {
                    //Make sure the box is drawn within the picture
                    int boxX = e.X, boxY = e.Y;
                    if (boxX < imgRect.Left)
                        boxX = imgRect.Left;
                    else if (boxX >= imgRect.Right)
                        boxX = imgRect.Right - 1;
                    if (boxY < imgRect.Top)
                        boxY = imgRect.Top;
                    else if (boxY >= imgRect.Bottom)
                        boxY = imgRect.Bottom - 1;

                    //g.Clear(pbInput.BackColor);
                    g.DrawImage(image, imgRect);
                    Point p1 = new Point(cutRect.X, cutRect.Y);
                    Point p2 = new Point(boxX, cutRect.Y);
                    Point p3 = new Point(cutRect.X, boxY);
                    Point p4 = new Point(boxX, boxY);
                    //Draw Box
                    g.DrawLine(pen, p1, p2);
                    g.DrawLine(pen, p2, p4);
                    g.DrawLine(pen, p4, p3);
                    g.DrawLine(pen, p3, p1);
                }
            }
        }

        private void pbInput_MouseUp(object sender, MouseEventArgs e)
        {
            //Finishing crop a picture
            if (isCut && e.Button == MouseButtons.Left && isDraw)
            {
                //if (e.X < imgRect.Left || e.X >= imgRect.Right || e.Y < imgRect.Top || e.Y >= imgRect.Bottom)
                //{
                //    MessageBox.Show("You have to cut on the picture.");
                //    isDraw = false;
                //    g.Clear(pbInput.BackColor);
                //    g.DrawImage(image, imgRect);
                //    return;
                //}

                //Make sure the box is drawn within the picture
                int boxX = e.X, boxY = e.Y;
                if (boxX < imgRect.Left)
                    boxX = imgRect.Left;
                else if (boxX >= imgRect.Right)
                    boxX = imgRect.Right - 1;
                if (boxY < imgRect.Top)
                    boxY = imgRect.Top;
                else if (boxY >= imgRect.Bottom)
                    boxY = imgRect.Bottom - 1;

                isDraw = false;
                isCut = false;
                cutRect.Width = boxX - cutRect.X;
                cutRect.Height = boxY - cutRect.Y;
            }
            //Pick a text color
            else if (isPick && e.Button == MouseButtons.Left)
            {
                //Tried to use an average color for the block around the picked point, but does not work good for the slim text.
                if (e.X < imgRect.Left || e.X >= imgRect.Right || e.Y < imgRect.Top || e.Y >= imgRect.Bottom)
                {
                    MessageBox.Show("You have to pick a color on the picture.");
                    isDraw = false;
                    return;
                }
                Pen pen1 = new Pen(Color.Orange, 2);
                g.DrawImage(image, imgRect);
                Point p1 = new Point(e.X - 2, e.Y - 2);
                Point p2 = new Point(e.X + 2, e.Y - 2);
                Point p3 = new Point(e.X - 2, e.Y + 2);
                Point p4 = new Point(e.X + 2, e.Y + 2);
                g.DrawLine(pen1, p1, p2);
                g.DrawLine(pen1, p2, p4);
                g.DrawLine(pen1, p4, p3);
                g.DrawLine(pen1, p3, p1);
                if (fitHeight)
                   textColor = image.GetPixel((int)((e.X - imgRect.X) / zoomrate), (int)(e.Y / zoomrate));              
                else textColor = image.GetPixel((int)(e.X / zoomrate), (int)((e.Y - imgRect.Y) / zoomrate));
                labelColor.BackColor = textColor;
            }
        }

        public Bitmap CropBitmap(Bitmap bitmap, int cropX, int cropY, int cropWidth, int cropHeight)
        {
            Rectangle rect = new Rectangle(cropX, cropY, cropWidth, cropHeight);
            System.Drawing.Imaging.PixelFormat format = bitmap.PixelFormat;
            Bitmap cropped = bitmap.Clone(rect, format);
            return cropped;
        }

        /********************************************************************
         * Function for Barinizing the picture, so black and white only     *
         * ******************************************************************/
        private void Binarize()
        {
            //Setup for Process Bar
            btFull.Enabled = false;
            tsLabel1.Text = "Binarizing: ";
            tsProgressBar.Visible = true;
            tsProgressBar.Value = 0;
            tsProgressBar.Minimum = 0;
            tsProgressBar.Maximum = image.Height;
            Refresh();

            for (int j = 0; j < image.Height; j++)            
            {
                for (int i = 0; i < image.Width; i++)
                {
                    Color c = image.GetPixel(i, j);
                    if (System.Math.Abs(c.R - textColor.R) + System.Math.Abs(c.G - textColor.G) + System.Math.Abs(c.B - textColor.B) <= 150)
                        image.SetPixel(i, j, Color.Black);
                    else image.SetPixel(i, j, Color.White);
                }
                tsProgressBar.Value++;
            }
            tsLabel1.Text = "Binarization Finished";
            tsProgressBar.Visible = false;
            btFull.Enabled = true;
        }

        /********************************************************************
         * Functions for Denoising                                          *
         * Any block is smaller than nBlockSize will be removed             *
         * Source:http://blog.csdn.net/stevenkylelee/article/details/6973471*
         * ******************************************************************/
        private void Denoising(Bitmap bm, int nBlockSize)
        {
            //Setup for Process Bar
            btFull.Enabled = false;
            tsLabel1.Text = "Denoising: ";
            tsProgressBar.Visible = true;
            tsProgressBar.Value = 0;
            tsProgressBar.Minimum = 0;
            tsProgressBar.Maximum = image.Width;
            Refresh();

            // Points that have been searched
            Dictionary<string, Point> Track = new Dictionary<string, Point>();
            for (int i = 0; i < bm.Width; ++i)
            {
                for (int j = 0; j < bm.Height; ++j)
                {
                    if (Track.ContainsKey(i + "#" + j) == true)
                        continue;

                    Dictionary<string, Point> Block = GetBlock(bm, i, j);
                    foreach (string strkey in Block.Keys)
                    {
                        Track.Add(strkey, Block[strkey]);
                    }

                    if (Block.Count < nBlockSize)
                    {
                        foreach (KeyValuePair<string, Point> Item in Block)
                        {
                            Point pt = Item.Value;
                            bm.SetPixel(pt.X, pt.Y, Color.White);
                        }
                    }
                }
                tsProgressBar.Value++;
            }
            tsLabel1.Text = "Denoising Finished";
            tsProgressBar.Visible = false;
            btFull.Enabled = true;
        }

        //This function is used to find the block, where the black points are linked together
        private static Dictionary<string, Point> GetBlock(Bitmap bm, int x, int y)
        {
            Dictionary<string, Point> Track = new Dictionary<string, Point>();
            string strKeyOfPoint;
            Stack<Point> stk = new Stack<Point>();
            Color Cr = bm.GetPixel(x, y);
            if (ArgbEqual(Cr, Color.White) == true)
            {
                return Track;
            }
            stk.Push(new Point(x, y));

            for (; stk.Count != 0; )
            {
                Point Pt = stk.Pop();
                strKeyOfPoint = Pt.X + "#" + Pt.Y;
                Track[strKeyOfPoint] = new Point(Pt.X, Pt.Y);
                List<Point> lstAdjacency = new List<Point>();

                // Right
                Point ptTest = new Point(Pt.X + 1, Pt.Y);
                if (ptTest.X < bm.Width)
                {
                    Color crTest = bm.GetPixel(ptTest.X, ptTest.Y);
                    if (ArgbEqual(crTest, Color.Black))
                    {
                        lstAdjacency.Add(ptTest);
                    }
                }

                // Left
                ptTest = new Point(Pt.X - 1, Pt.Y);
                if (ptTest.X >= 0)
                {
                    Color crTest = bm.GetPixel(ptTest.X, ptTest.Y);
                    if (ArgbEqual(crTest, Color.Black))
                    {
                        lstAdjacency.Add(ptTest);
                    }
                }

                // Down
                ptTest = new Point(Pt.X, Pt.Y + 1);
                if (ptTest.Y < bm.Height)
                {
                    Color crTest = bm.GetPixel(ptTest.X, ptTest.Y);
                    if (ArgbEqual(crTest, Color.Black))
                    {
                        lstAdjacency.Add(ptTest);
                    }
                }

                // Up
                ptTest = new Point(Pt.X, Pt.Y - 1);
                if (ptTest.Y >= 0)
                {
                    Color crTest = bm.GetPixel(ptTest.X, ptTest.Y);
                    if (ArgbEqual(crTest, Color.Black))
                    {
                        lstAdjacency.Add(ptTest);
                    }
                }

                for (int i = 0; i < lstAdjacency.Count; ++i)
                {
                    Point ptAdjacency = lstAdjacency[i];
                    strKeyOfPoint = ptAdjacency.X + "#" + ptAdjacency.Y;
                    if (Track.ContainsKey(strKeyOfPoint) == false)
                    {
                        stk.Push(ptAdjacency);
                    }
                }
            }
            return Track;
        }

        /*******************************************************************
         * Function for Recognition Text                                   *
         * *****************************************************************/
        private void RecognitionText()
        {
            //Setup for Process Bar
            btFull.Enabled = false;
            tsLabel1.Text = "Recognizing: ";
            tsProgressBar.Visible = true;
            tsProgressBar.Value = 0;
            tsProgressBar.Minimum = 0;
            tsProgressBar.Maximum = 10000;
            Refresh();

            int k = 0;
            int pixelCount = 0;
            int emptyLineCount = 0;
            int writtenLineCount = 0;
            bool empty = true;
            tbOutput.Text = "";
            tessnet2.Tesseract ocr = new tessnet2.Tesseract();
            string charList = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-.,?()\\;:'!@_=+#$%&*";
            ocr.SetVariable("tessedit_char_whitelist", charList); // If digit only
            ocr.Init(Application.StartupPath + @"\\tessdata", "eng", false); // English Only
            List<tessnet2.Word> result;

            //Cutting Lines
            for (int i = 0; i < image.Width; i++)
            {
                Color c = image.GetPixel(i, 0);
                if (ArgbEqual(c, Color.Black))
                    pixelCount++;
            }
            if (pixelCount > emptyLineCoefficient * image.Width)   //Not empty line
                empty = false;
            for (int j = 0; j < image.Height; j++)
            {
                if(k>=1000)     //Because the array has size 1000, so the program cannot handle a picture with more than 1000 lines of text
                {
                    MessageBox.Show("Cannot have more than 1000 lines of text./r/nPlease choose a smaller picture.");
                    newToolStripButton.PerformClick();
                    return;
                }
                pixelCount = 0;
                for (int i = 0; i < image.Width; i++)
                {
                    Color c = image.GetPixel(i, j);
                    if (ArgbEqual(c, Color.Black))
                        pixelCount++;
                }
                if (pixelCount <= emptyLineCoefficient * image.Width && !empty)   //Find Empty line when pervious is not
                {
                    writtenLineCount = 0;
                    if (emptyLineCount < 3)         //Three continuous empty lines counts as a gap between 2 lines of text
                        emptyLineCount++;
                    else
                    {
                        emptyLineCount = 0;
                        lineCut[k] = j;
                        empty = true;
                        k++;
                    }
                }
                else if (pixelCount > emptyLineCoefficient * image.Width && empty) //Find written line when pervious is empty
                {
                    emptyLineCount = 0;
                    if (writtenLineCount < 3)        //Three continuous non-empty lines counts as the sign of text
                        writtenLineCount++;
                    else
                    {
                        writtenLineCount = 0;
                        empty = false;
                    }
                }
                tsProgressBar.Value += tsProgressBar.Maximum / 2 / image.Height;
            }
            if (!empty)                 //If the last line is not empty, then count it!
            {
                lineCut[k] = image.Height - 1;
                k++;
            }
            if (k == 0)           //if k never used, this is a empty image after binarizion.
            {
                MessageBox.Show("Opps...  I can't see any text, you may want to try again with the Full Process.");
                return;
            }

            //Run OCR engine
            result = ocr.DoOCR(image, new Rectangle(0, 0, image.Width, lineCut[0]));
            foreach (tessnet2.Word word in result)
                if (word.Text != "~")
                    tbOutput.Text += (word.Text + " ");
            tbOutput.Text += "\r\n";
            for (int i = 0; i < k - 1; i++)
            {
                result = ocr.DoOCR(image, new Rectangle(0, lineCut[i], image.Width, lineCut[i + 1] - lineCut[i]));
                foreach (tessnet2.Word word in result)
                    if (word.Text != "~")
                        tbOutput.Text += (word.Text + " ");
                tbOutput.Text += "\r\n";
                tsProgressBar.Value += tsProgressBar.Maximum / 2 / k;
            }
            tsLabel1.Text = "Recognition Finished";
            tsProgressBar.Visible = false;
            btFull.Enabled = true;
            MessageBox.Show("Recognition Finished.\r\nSee result in the right text window.");
        }

        /********************************************************************
         * Some useful samll functions                                      *
         * ******************************************************************/
        public bool IsNumeric(string str)
        {
            foreach (char ch in str)
                if (!char.IsNumber(ch))
                    return false;
            return true;
        }

        private static bool ArgbEqual(Color cr1, Color cr2)
        {
            if (cr1.A == cr2.A && cr1.R == cr2.R && cr1.G == cr2.G && cr1.B == cr2.B)
                return true;
            else return false;
        }
    }
}
